"""
NATS integration for taskiq.

This package contains brokers that
uses NATS as a message queue.
"""

from taskiq_nats.broker import (
    NatsBroker,
    PullBasedJetStreamBroker,
    PushBasedJetStreamBroker,
)
from taskiq_nats.result_backend import NATSObjectStoreResultBackend
from taskiq_nats.schedule_source import NATSKeyValueScheduleSource

__all__ = [
    "NatsBroker",
    "PushBasedJetStreamBroker",
    "PullBasedJetStreamBroker",
    "NATSObjectStoreResultBackend",
    "NATSKeyValueScheduleSource",
]



import typing
from abc import ABC, abstractmethod
from logging import getLogger

from nats.aio.client import Client
from nats.aio.msg import Msg as NatsMessage
from nats.errors import TimeoutError as NatsTimeoutError
from nats.js import JetStreamContext
from nats.js.api import ConsumerConfig, StreamConfig
from taskiq import AckableMessage, AsyncBroker, AsyncResultBackend, BrokerMessage

_T = typing.TypeVar("_T")  # (Too short)


JetStreamConsumerType = typing.TypeVar(
    "JetStreamConsumerType",
)


logger = getLogger("taskiq_nats")


class NatsBroker(AsyncBroker):
    """
    NATS broker for taskiq.

    By default this broker works
    broadcasting message to all connected workers.

    If you want to make it work as queue,
    you need to supply name of the queue in
    queue argument.

    Docs about queue:
    https://docs.nats.io/nats-concepts/core-nats/queue
    """

    def __init__(
        self,
        servers: typing.Union[str, typing.List[str]],
        subject: str = "taskiq_tasks",
        queue: typing.Optional[str] = None,
        result_backend: "typing.Optional[AsyncResultBackend[_T]]" = None,
        task_id_generator: typing.Optional[typing.Callable[[], str]] = None,
        **connection_kwargs: typing.Any,
    ) -> None:
        super().__init__(result_backend, task_id_generator)
        self.servers = servers
        self.client: Client = Client()
        self.connection_kwargs = connection_kwargs
        self.queue = queue
        self.subject = subject

    async def startup(self) -> None:
        """
        Startup event handler.

        It simply connects to NATS cluster.
        """
        await super().startup()
        await self.client.connect(self.servers, **self.connection_kwargs)

    async def kick(self, message: BrokerMessage) -> None:
        """
        Send a message using NATS.

        :param message: message to send.
        """
        await self.client.publish(
            self.subject,
            payload=message.message,
            headers=message.labels,
        )

    async def listen(self) -> typing.AsyncGenerator[bytes, None]:
        """
        Start listen to new messages.

        :yield: incoming messages.
        """
        subscribe = await self.client.subscribe(self.subject, queue=self.queue or "")
        async for message in subscribe.messages:
            yield message.data

    async def shutdown(self) -> None:
        """Close connections to NATS."""
        await self.client.close()
        await super().shutdown()


class BaseJetStreamBroker(
    AsyncBroker,
    ABC,
    typing.Generic[JetStreamConsumerType],
):
    """Base JetStream broker for taskiq.

    It has two subclasses - PullBasedJetStreamBroker
    and PushBasedJetStreamBroker.

    These brokers create a JetStream context
    and use it to send and receive messages.

    This is useful for systems where you need to
    be sure that messages are delivered to the workers.
    """

    def __init__(
        self,
        servers: typing.Union[str, typing.List[str]],
        subject: str = "taskiq_tasks",
        stream_name: str = "taskiq_jetstream",
        queue: typing.Optional[str] = None,
        durable: str = "taskiq_durable",
        stream_config: typing.Optional[StreamConfig] = None,
        consumer_config: typing.Optional[ConsumerConfig] = None,
        pull_consume_batch: int = 1,
        pull_consume_timeout: typing.Optional[float] = None,
        **connection_kwargs: typing.Any,
    ) -> None:
        super().__init__()
        self.servers: typing.Final = servers
        self.client: typing.Final = Client()
        self.connection_kwargs: typing.Final = connection_kwargs
        self.subject: typing.Final = subject
        self.stream_name: typing.Final = stream_name
        self.js: JetStreamContext
        self.stream_config = stream_config or StreamConfig()
        self.consumer_config = consumer_config

        # Only for push based consumer
        self.queue: typing.Final = queue
        self.default_consumer_name: typing.Final = "taskiq_consumer"
        # Only for pull based consumer
        self.durable: typing.Final = durable
        self.pull_consume_batch: typing.Final = pull_consume_batch
        self.pull_consume_timeout: typing.Final = pull_consume_timeout

        self.consumer: JetStreamConsumerType

    async def startup(self) -> None:
        """
        Startup event handler.

        It simply connects to NATS cluster, and
        setup JetStream.
        """
        await super().startup()
        await self.client.connect(self.servers, **self.connection_kwargs)
        self.js = self.client.jetstream()
        if self.stream_config.name is None:
            self.stream_config.name = self.stream_name
        if not self.stream_config.subjects:
            self.stream_config.subjects = [self.subject]
        await self.js.add_stream(config=self.stream_config)
        await self._startup_consumer()

    async def shutdown(self) -> None:
        """Close connections to NATS."""
        await self.client.close()
        await super().shutdown()

    async def kick(self, message: BrokerMessage) -> None:
        """
        Send a message using NATS.

        :param message: message to send.
        """
        await self.js.publish(
            self.subject,
            payload=message.message,
            headers=message.labels,
        )

    @abstractmethod
    async def _startup_consumer(self) -> None:
        """Create consumer."""


class PushBasedJetStreamBroker(
    BaseJetStreamBroker[JetStreamContext.PushSubscription],
):
    """JetStream broker for push based message consumption.

    It's named `push` based because nats server push messages to
    the consumer, not consumer requests them.
    """

    async def listen(self) -> typing.AsyncGenerator[AckableMessage, None]:
        """
        Start listen to new messages.

        :yield: incoming messages.
        """
        async for message in self.consumer.messages:
            yield AckableMessage(
                data=message.data,
                ack=message.ack,
            )

    async def _startup_consumer(self) -> None:
        if not self.consumer_config:
            self.consumer_config = ConsumerConfig(
                name=self.default_consumer_name,
                durable_name=self.default_consumer_name,
            )

        self.consumer = await self.js.subscribe(
            subject=self.subject,
            queue=self.queue or "",
            config=self.consumer_config,
        )


class PullBasedJetStreamBroker(
    BaseJetStreamBroker[JetStreamContext.PullSubscription],
):
    """JetStream broker for pull based message consumption.

    It's named `pull` based because consumer requests messages,
    not NATS server sends them.
    """

    async def listen(self) -> typing.AsyncGenerator[AckableMessage, None]:
        """
        Start listen to new messages.

        :yield: incoming messages.
        """
        while True:
            try:
                nats_messages: typing.List[NatsMessage] = await self.consumer.fetch(
                    batch=self.pull_consume_batch,
                    timeout=self.pull_consume_timeout,
                )
                for nats_message in nats_messages:
                    yield AckableMessage(
                        data=nats_message.data,
                        ack=nats_message.ack,
                    )
            except NatsTimeoutError:
                continue

    async def _startup_consumer(self) -> None:
        if not self.consumer_config:
            self.consumer_config = ConsumerConfig(
                durable_name=self.durable,
            )

        # We must use this method to create pull based consumer
        # because consumer config won't change without it.
        await self.js.add_consumer(
            stream=self.stream_config.name or self.stream_name,
            config=self.consumer_config,
        )
        self.consumer = await self.js.pull_subscribe(
            subject=self.subject,
            durable=self.durable,
            config=self.consumer_config,
        )



import logging
from typing import Any, Final, List, Optional, Union

import nats
from nats import NATS
from nats.js import JetStreamContext
from nats.js.errors import BucketNotFoundError, NoKeysError
from nats.js.kv import KeyValue
from taskiq import ScheduledTask, ScheduleSource
from taskiq.abc.serializer import TaskiqSerializer
from taskiq.compat import model_dump, model_validate
from taskiq.serializers import PickleSerializer

log = logging.getLogger(__name__)


class NATSKeyValueScheduleSource(ScheduleSource):
    """
    Source of schedules for NATS Key-Value storage.

    This class allows you to store schedules in NATS Key-Value storage.
    Also it supports dynamic schedules.
    """

    def __init__(
        self,
        servers: Union[str, List[str]],
        bucket_name: str = "taskiq_schedules",
        prefix: str = "schedule",
        serializer: Optional[TaskiqSerializer] = None,
        **connect_options: Any,
    ) -> None:
        """Construct new result backend.

        :param servers: NATS servers.
        :param bucket_name: name of the bucket where schedules would be stored.
        :param prefix: prefix for nats kv storage schedule keys.
        :param serializer: serializer for data.
        :param connect_kwargs: additional arguments for nats `connect()` method.
        """
        self.servers: Final = servers
        self.bucket_name: Final = bucket_name
        self.prefix: Final = prefix
        self.serializer = serializer or PickleSerializer()
        self.connect_options: Final = connect_options

        self.nats_client: NATS
        self.nats_jetstream: JetStreamContext
        self.kv: KeyValue

    async def startup(self) -> None:
        """Create new connection to NATS.

        Initialize JetStream context and new KeyValue instance.
        """
        self.nats_client = await nats.connect(
            servers=self.servers,
            **self.connect_options,
        )
        self.nats_jetstream = self.nats_client.jetstream()

        try:
            self.kv = await self.nats_jetstream.key_value(self.bucket_name)
        except BucketNotFoundError:
            self.kv = await self.nats_jetstream.create_key_value(
                bucket=self.bucket_name,
            )

    async def shutdown(self) -> None:
        """Close nats connection."""
        if self.nats_client.is_closed:
            return
        await self.nats_client.close()

    async def delete_schedule(self, schedule_id: str) -> None:
        """Remove schedule by id."""
        await self.kv.delete(f"{self.prefix}.{schedule_id}")

    async def add_schedule(self, schedule: ScheduledTask) -> None:
        """
        Add schedule to NATS Key-Value storage.

        :param schedule: schedule to add.
        :param schedule_id: schedule id.
        """
        await self.kv.put(
            f"{self.prefix}.{schedule.schedule_id}",
            self.serializer.dumpb(model_dump(schedule)),
        )

    async def get_schedules(self) -> List[ScheduledTask]:
        """
        Get all schedules from NATS Key-Value storage.

        This method is used by scheduler to get all schedules.

        :return: list of schedules.
        """
        try:
            schedules = await self.kv.history(f"{self.prefix}.*")
        except NoKeysError:
            return []

        return [
            model_validate(ScheduledTask, self.serializer.loadb(schedule.value))
            for schedule in schedules
            if schedule and schedule.value
        ]

    async def post_send(self, task: ScheduledTask) -> None:
        """Delete a task after it's completed."""
        if task.time is not None:
            await self.delete_schedule(task.schedule_id)



from typing import Any, Final, List, Optional, TypeVar, Union

import nats
from nats.aio.client import Client
from nats.js import JetStreamContext
from nats.js.errors import BucketNotFoundError, ObjectNotFoundError
from nats.js.object_store import ObjectStore
from taskiq import AsyncResultBackend, ResultGetError
from taskiq.abc.serializer import TaskiqSerializer
from taskiq.compat import model_dump, model_validate
from taskiq.result import TaskiqResult
from taskiq.serializers import PickleSerializer

_ReturnType = TypeVar("_ReturnType")


class NATSObjectStoreResultBackend(AsyncResultBackend[_ReturnType]):
    """Result backend for NATS Object Store."""

    def __init__(
        self,
        servers: Union[str, List[str]],
        keep_results: bool = True,
        bucket_name: str = "taskiq_results",
        serializer: Optional[TaskiqSerializer] = None,
        **connect_options: Any,
    ) -> None:
        """Construct new result backend.

        :param servers: NATS servers .
        :param keep_results: flag to not remove results from Redis after reading.
        :param connect_kwargs: additional arguments for nats `connect()` method.
        """
        self.servers: Final = servers
        self.keep_results: Final = keep_results
        self.bucket_name: Final = bucket_name
        self.serializer = serializer or PickleSerializer()
        self.connect_options: Final = connect_options

        self.nats_client: Client
        self.nats_jetstream: JetStreamContext
        self.object_store: ObjectStore

    async def startup(self) -> None:
        """Create new connection to NATS.

        Initialize JetStream context and new ObjectStore instance.
        """
        self.nats_client = await nats.connect(
            servers=self.servers,
            **self.connect_options,
        )
        self.nats_jetstream = self.nats_client.jetstream()

        try:
            self.object_store = await self.nats_jetstream.object_store(self.bucket_name)
        except BucketNotFoundError:
            self.object_store = await self.nats_jetstream.create_object_store(
                self.bucket_name,
            )

    async def shutdown(self) -> None:
        """Close nats connection."""
        if self.nats_client.is_closed:
            return
        await self.nats_client.close()

    async def set_result(self, task_id: str, result: TaskiqResult[_ReturnType]) -> None:
        """Set result to the nats bucket.

        :param task_id: ID of the task.
        :param result: result of the task.
        """
        await self.object_store.put(
            name=task_id,
            data=self.serializer.dumpb(model_dump(result)),
        )

    async def is_result_ready(self, task_id: str) -> bool:
        """Returns whether the result is ready.

        :param task_id: ID of the task.

        :returns: True if the result is ready else False.
        """
        try:
            await self.object_store.get(name=task_id)
        except ObjectNotFoundError:
            return False
        return True

    async def get_result(
        self,
        task_id: str,
        with_logs: bool = False,
    ) -> TaskiqResult[_ReturnType]:
        """
        Retrieve result from the task.

        :param task_id: task's id.
        :param with_logs: if True it will download task's logs.
        :raises ResultIsMissingError: if there is no result when trying to get it.
        :return: TaskiqResult.
        """
        try:
            result = await self.object_store.get(
                name=task_id,
            )
        except ObjectNotFoundError as exc:
            raise ResultGetError from exc

        if not self.keep_results:
            await self.object_store.delete(
                name=task_id,
            )

        taskiq_result: TaskiqResult[_ReturnType] = model_validate(
            TaskiqResult[_ReturnType],  # type: ignore[misc]
            self.serializer.loadb(result.data),
        )

        if not with_logs:
            taskiq_result.log = None

        return taskiq_result
